package it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.model;

import java.util.HashSet;
import java.util.Set;

import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.logging.ModelEventLogger;
import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.model.CalculatorEnableEvent.Target;
import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.view.CalculatorView;

/**
 * Un modello di calcolatrice semplificata. 
 */
public class CalculatorModel {

  /* ultimo risultato */  
  private double lastResult;

  /* ultimo operando ed operatore selezionati */
  private double currentOperand;
  private Operation currentOperation;

  /* stato della calcolatrice */
  private boolean hasOperand;
  private boolean hasOperation;

  /* Observer: Subject. 
   * rappresenta l'insieme di viste
   * registrate al modello. */
  private Set<CalculatorView> registeredViews;

  /* logger degli eventi generati dal modello */
  private final ModelEventLogger logger;
  private final boolean hasLogger;
  
  /**
   * Crea un nuovo modello di calcolatrice.
   * @param logger Un logger per gli eventi generati dal modello.
   */
  public CalculatorModel(ModelEventLogger logger) {
    this.registeredViews = new HashSet<CalculatorView>();
    this.logger = logger;
    this.hasLogger = (logger != null);
    this.reset();
  }
  
  /**
   * Crea un nuovo modello di calcolatrice.
   */
  public CalculatorModel() {
    this(null);
  }
  

  /**
   * Registra una User Interface .
   * @param view una interfaccia utente come vista per 
   * il modello di calcolatrice.
   */
  public void addView(CalculatorView view) {
    /* salvo riferimento alla UI */
    this.registeredViews.add(view);
    /* mi porto allo stato iniziale */
    this.reset();
  }
  
  /**
   * Annulla la registrazione di una vista al modello
   * di calcolatrice.
   * @param view una vista.
   */
  public void detachView(CalculatorView view) {
    this.registeredViews.remove(view);
  }

  /**
   * Notifica alle viste l'occorrenza di un evento.
   * @param event un evento da notificare alle viste.
   */
  private void notifyObservers(CalculatorModelEvent event) {
    for (CalculatorView view: this.registeredViews) {
      view.notify(event);
    }
  }
  
  /**
   * Imposta la calcolatrice nello stato iniziale.
   */
  public void reset() {
    this.lastResult = 0.0;

    /* imposta lo stato iniziale */
    this.hasOperand = false;
    this.hasOperation = false;

    /* imposta valori iniziali */
    this.currentOperand = 0.0;
    this.currentOperation = null;
    
    generareAndDispatchEventsForViews(0.0, false, false);
  }

  /**
   * Imposta l'operazione corrente
   * @param op operazione corrente.
   */
  public void setCurrentOperation(Operation op) {
    if (this.hasOperand)
      compute();

    this.hasOperand = false;
    this.hasOperation = true;

    this.currentOperation = op;

    this.generareAndDispatchEventsForViews(this.lastResult, true, true);
  }

  /**
   * Imposta l'operando corrente
   * @param value l'operando corrente
   */
  public void setOperand(double value) {
    this.hasOperand = true;
    this.currentOperand = value;
  }
  
  
  /**
   * Aggiorna il risultato usando operazione e operando corrente.
   * 
   * Questo metodo deve essere chiamato solo se l'operazione e' stata selezionata prima.
   * Se non e' stato selezionato un operando, viene utilizzato di default il risultato
   * della precedente operazione. 
   * 
   * @throws IllegalStateException se non e' stato selezionata una operazione.
   */
  public void compute() {
    if (!this.hasOperation)
      throw new IllegalStateException();

    double argument = 0.0;
    if (this.hasOperand)
      argument = this.currentOperand;
    else 
      argument = this.lastResult;

    this.lastResult = 
        this.currentOperation.perform(this.lastResult, argument);

    this.hasOperand = false;
    this.hasOperation = false;

    this.generareAndDispatchEventsForViews(this.lastResult, false, false);
  }
  
  private void generareAndDispatchEventsForViews(double result, boolean enableOperand, boolean enableEquals) {
    
    CalculatorModelEvent resultEvent = 
        CalculatorModelEventFactory.resultEvent(result),
          digitEvent = 
        CalculatorModelEventFactory.enableEvent(Target.OPERAND, enableOperand),
          equalsEvent = 
        CalculatorModelEventFactory.enableEvent(Target.EQUALS, enableEquals);

    dispatchEventsForViews(resultEvent, digitEvent, equalsEvent);
  }
  
  private void dispatchEventsForViews(CalculatorModelEvent... events) {
    if (events == null)
      throw new IllegalArgumentException("invalid list of events, can't be null");
    
    for(CalculatorModelEvent event: events) {
      if (event == null)
        throw new IllegalArgumentException("can't dispatch a null event");
      
      if (this.hasLogger)
        this.logger.log(event);
      
      notifyObservers(event);
    }
  }
}
