package it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.view;

import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.handlers.ControllerHandlerFactory;
import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.handlers.GuiHandlerFactory;
import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.model.CalculatorEnableEvent;
import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.model.CalculatorEnableEvent.Target;
import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.model.CalculatorModelEvent;
import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.model.CalculatorResultEvent;
import it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator.model.Operation;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

/**
 * Rappresenta una Interfaccia utente grafica per 
 * una calcolatrice.
 * @author andrea
 *
 */
public class CalculatorGUI implements CalculatorView {

  // finestra principale
  private final JFrame calculatorFrame;
  
  // rappresenta il risultato
  private JLabel result;
  // rappresenta l'operando corrente
  private JTextField value;

  // lista di bottoni per le cifre
  private List<JButton> digitsButtons;
  // lista di bottoni per le operazioni
  private List<JButton> operationButtons;
  // altri bottoni: virgola, annulla, =
  private JButton dotButton, cancelButton, equalsButton;

  private final ControllerHandlerFactory controllerHandlerFactory;
  private final GuiHandlerFactory guiHandlerFactory;

  /**
   * Crea una nuova GUI.
   * @param math Un modello di calcolatrice.
   */
  public CalculatorGUI(ControllerHandlerFactory controllerHandlerFactory) {
    this.calculatorFrame = new JFrame("Simple Calculator");
    this.controllerHandlerFactory = controllerHandlerFactory;
    this.guiHandlerFactory = new GuiHandlerFactory(this);
    setup();
  }
  
  // ==============================================================
  // Questa classe ha tre parti:
  // 1. Metodi privati per la costruzione della gui,
  // 2. Metodi pubblici chiamati dal modello per aggiornare la gui
  // 3. Classi interne che rappresentano gli handler per gestire
  //    gli eventi della GUI.
  // ==============================================================
  
  // ==============================================================
  // METODI PRIVATI PER LA COSTRUZIONE DELLA GUI
  // ==============================================================

  // Definisce la struttura del panel del frame principale
  // e ne costruisce le sue parti
  private void setup() {
    setupButtons();

    this.calculatorFrame.setLayout(new BorderLayout());

    JPanel upperPanel = buildUpperPanel();
    this.calculatorFrame.add(upperPanel, BorderLayout.NORTH);

    JPanel centralPanel = buildCentralPanel();
    this.calculatorFrame.add(centralPanel, BorderLayout.CENTER);

    JPanel operationsPanel = buildOperationsPanel();
    this.calculatorFrame.add(operationsPanel, BorderLayout.EAST);

    this.calculatorFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    this.calculatorFrame.pack();
    this.calculatorFrame.setVisible(true);
    // lo rendo non resizable per evitare dimensioni in cui i componenti
    // non sono piu' visibili.
    this.calculatorFrame.setResizable(false);
  }

  // costruisce il pannello sulle operazioni
  private JPanel buildOperationsPanel() {
    JPanel operationsPanel = new JPanel(new GridLayout(5, 1));

    for (JButton operationButton: this.operationButtons) {
      operationsPanel.add(operationButton);
    }

    operationsPanel.add(equalsButton);
    return operationsPanel;
  }

  // costruisce il pannello centrale con le operazioni
  private JPanel buildCentralPanel() {
    JPanel centralPanel = new JPanel(new GridLayout(4, 3));
    for (JButton digitButton: this.digitsButtons) {
      centralPanel.add(digitButton);
    }

    centralPanel.add(dotButton);
    centralPanel.add(this.cancelButton);
    return centralPanel;
  }

  // costruisce il pannello superiore con label del risulato
  // e textfield dell'operando
  private JPanel buildUpperPanel() {
    JPanel upperPanel = new JPanel(new GridLayout(2, 1));
    result = new JLabel("0.0");
    upperPanel.add(result);
    value = new JTextField("");

    // Non deve essere editabile direttamente dall'utente
    value.setEditable(false); 
    upperPanel.add(value);
    return upperPanel;
  }

  private void setupButtons() {
    // crea i bottoni per le cifre
    createDigitButtons();
    // crea i bottoni per le operazioni
    createOperationButtons();

    this.cancelButton = new JButton("C");
    this.cancelButton.addActionListener(this.controllerHandlerFactory.cancelButtonHandler());
  }

  
  
  // crea i bottoni corrispondenti alle cifre e alla virgola
  private void createDigitButtons() {
    this.digitsButtons = new ArrayList<>(10);

    ActionListener digitButtonControllerHandler = 
        this.controllerHandlerFactory.digitButtonHandler();
    
    for (int i = 1; i <= 10; i++) {
      int digit = i % 10;
      JButton digitButton = new JButton(String.valueOf(digit));
      /* ogni cifra ha bisogno di due handler:
       * uno e' relativo alla sola GUI, per cambiare lo stato
       * della casella del valore, l'altro e' relativo
       * al controller, per settare l'operando nel modello. 
       * I due devono essere
       * eseguiti in un certo ordine, quindi li combino in un 
       * modo un po' particolare (analoghe soluzioni sono possibili)
       */
      ActionListener digitButtonHandler = 
          guiHandlerFactory.digitButtonHandler(
              digitButtonControllerHandler,
              digit);
      digitButton.addActionListener(digitButtonHandler);
      this.digitsButtons.add(digitButton);
    }

    this.dotButton = new JButton(".");
    this.dotButton.addActionListener(this.guiHandlerFactory.dotButtonHandler());
  }

  // crea i bottoni corrispondenti alle operazioni, compresa l'uguaglianza.
  private void createOperationButtons() {
    this.operationButtons = new ArrayList<>(Operation.values().length);

    for (Operation operation: Operation.values()) {
      /* non usare -name()- per le enum, ma implementare la toString, come in questo caso */
      JButton operationButton = new JButton(operation.toString());
      ActionListener operationButtonHandler = controllerHandlerFactory.operationButtonHandler(operation);
      operationButton.addActionListener(operationButtonHandler);
      this.operationButtons.add(operationButton);
    }		

    this.equalsButton = new JButton("=");
    this.equalsButton.addActionListener(controllerHandlerFactory.equalsButtonHandler());
  }
  
  // ==============================================================
  //  Metodi di gestione degli eventi generati dal modello
  //  La view e' sottoscritta agli eventi di cambiamento
  //  di stato del modello
  // ==============================================================
  
  private void process(CalculatorResultEvent event) {
    double value = event.getResult();
    this.result.setText(String.valueOf(value));  
    this.value.setText("");
  }

  private void process(CalculatorEnableEvent event) {
    Target target = event.getTarget();
    boolean enabled = event.isEnabled();
    
    switch(target) {
    case OPERAND:
      setDigits(enabled); 
      break;
    case EQUALS:
      setEquals(enabled);
      break;
    default:
      /* ignore */
      break;
    }
  }
  
  private void setDigits(boolean enabled) {
    for (JButton digit: this.digitsButtons) {
      digit.setEnabled(enabled);
    }
    this.dotButton.setEnabled(enabled);
  }
  
  private void setEquals(boolean enabled) {
    this.equalsButton.setEnabled(enabled);
  }
  
  @Override
  public void notify(CalculatorModelEvent event) {
    if (event instanceof CalculatorResultEvent)
      process((CalculatorResultEvent)event);
    if  (event instanceof CalculatorEnableEvent)
      process((CalculatorEnableEvent)event);
    
    /* ignore other events */
  }

  // ==============================================================
  //  Metodi utilizzati dal controller e dagli handler degli eventi
  //  per la sola GUI
  // ==============================================================

  public String getOperandText() {
    return this.value.getText();
  }
  
  public void addDot() {
    if (!value.getText().contains("."))
      value.setText(value.getText() + ".");
  }
  
  public void addDigit(int digit) {
    value.setText(value.getText() + String.valueOf(digit));
  }

  @Override
  public String getOperandRepresentation() {
    return value.getText();
  }
}
